@@ -1,4 +1,4 @@
-# Copyright (c) 2013-2017 Christian Geier et al.
+# Copyright (c) 2013-2022 khal contributors
 #
 # Permission is hereby granted, free of charge, to any person obtaining
 # a copy of this software and associated documentation files (the
@@ -30,51 +30,57 @@ import itertools
 import logging
 import os
 import os.path
-from typing import Any, Dict, Iterable, List, Optional, Tuple, Union  # noqa
+from typing import Any, Dict, Iterable, List, Optional, Set, Tuple, Union  # noqa
 
+from ..custom_types import CalendarConfiguration, EventCreationTypes
+from ..icalendar import new_vevent
 from . import backend
 from .event import Event
-from .exceptions import (CouldNotCreateDbDir, DuplicateUid, NonUniqueUID,
-                         ReadOnlyCalendarError, UnsupportedFeatureError,
-                         UpdateFailed)
-from .vdir import (AlreadyExistingError, CollectionNotFoundError, Vdir,
-                   get_etag_from_file)
+from .exceptions import (
+    DuplicateUid,
+    EtagMissmatch,
+    NonUniqueUID,
+    ReadOnlyCalendarError,
+    UnsupportedFeatureError,
+    UpdateFailed,
+)
+from .vdir import (
+    AlreadyExistingError,
+    CollectionNotFoundError,
+    Vdir,
+    WrongEtagError,
+    get_etag_from_file,
+)
 
 logger = logging.getLogger('khal')
 
 
-def create_directory(path: str):
-    if not os.path.isdir(path):
-        if os.path.exists(path):
-            raise RuntimeError('{0} is not a directory.'.format(path))
-        try:
-            os.makedirs(path, mode=0o750)
-        except OSError as error:
-            logger.critical('failed to create {0}: {1}'.format(path, error))
-            raise CouldNotCreateDbDir()
-
-
-class CalendarCollection(object):
+class CalendarCollection:
     """CalendarCollection allows access to various calendars stored in vdirs
 
     all calendars are cached in an sqlitedb for performance reasons"""
 
     def __init__(self,
-                 calendars=None,
+                 calendars: Dict[str, CalendarConfiguration],
                  hmethod: str='fg',
                  default_color: str='',
                  multiple: str='',
+                 multiple_on_overflow: bool=False,
                  color: str='',
                  priority: int=10,
                  highlight_event_days: bool=False,
-                 locale: Dict[str, Any]=dict(),
+                 locale: Optional[Dict[str, Any]]=None,
                  dbpath: Optional[str]=None,
                  ) -> None:
+        locale = locale or {}
         assert dbpath is not None
         assert calendars is not None
-        self._calendars = calendars
-        self._default_calendar_name = None  # type: Optional[str]
-        self._storages = dict()  # type: Dict[str, Vdir]
+
+        self._calendars: Dict[str, CalendarConfiguration] = calendars
+        self._default_calendar_name: Optional[str] = None
+        self._storages: Dict[str, Vdir] = {}
+        file_ext: str
+
         for name, calendar in self._calendars.items():
             ctype = calendar.get('ctype', 'calendar')
             if ctype == 'calendar':
@@ -87,18 +93,19 @@ class CalendarCollection(object):
                 self._storages[name] = Vdir(calendar['path'], file_ext)
             except CollectionNotFoundError:
                 os.makedirs(calendar['path'])
-                logger.info('created non-existing vdir {}'.format(calendar['path']))
+                logger.info(f"created non-existing vdir {calendar['path']}")
                 self._storages[name] = Vdir(calendar['path'], file_ext)
 
         self.hmethod = hmethod
         self.default_color = default_color
         self.multiple = multiple
+        self.multiple_on_overflow = multiple_on_overflow
         self.color = color
         self.priority = priority
         self.highlight_event_days = highlight_event_days
         self._locale = locale
         self._backend = backend.SQLiteDb(self.names, dbpath, self._locale)
-        self._last_ctags = dict()  # type: Dict[str, str]
+        self._last_ctags: Dict[str, str] = {}
         self.update_db()
 
     @property
@@ -106,7 +113,7 @@ class CalendarCollection(object):
         return [c for c in self._calendars if not self._calendars[c].get('readonly', False)]
 
     @property
-    def calendars(self) -> Iterable[str]:
+    def calendars(self) -> Iterable[CalendarConfiguration]:
         return self._calendars.values()
 
     @property
@@ -118,11 +125,11 @@ class CalendarCollection(object):
         return self._default_calendar_name
 
     @default_calendar_name.setter
-    def default_calendar_name(self, default: str):
+    def default_calendar_name(self, default: str) -> None:
         if default is None:
             self._default_calendar_name = default
         elif default not in self.names:
-            raise ValueError('Unknown calendar: {0}'.format(default))
+            raise ValueError(f'Unknown calendar: {default}')
 
         readonly = self._calendars[default].get('readonly', False)
 
@@ -130,7 +137,7 @@ class CalendarCollection(object):
             self._default_calendar_name = default
         else:
             raise ValueError(
-                'Calendar "{0}" is read-only and cannot be used as default'.format(default))
+                f'Calendar "{default}" is read-only and cannot be used as default')
 
     def _local_ctag(self, calendar: str) -> str:
         return get_etag_from_file(self._calendars[calendar]['path'])
@@ -150,7 +157,7 @@ class CalendarCollection(object):
         floating_events = self.get_floating(start, end)
         localize = self._locale['local_timezone'].localize
         localized_events = self.get_localized(localize(start), localize(end))
-        return itertools.chain(floating_events, localized_events)
+        return itertools.chain(localized_events, floating_events)
 
     def get_calendars_on(self, day: dt.date) -> List[str]:
         start = dt.datetime.combine(day, dt.time.min)
@@ -162,9 +169,12 @@ class CalendarCollection(object):
         )
         return list(set(calendars))
 
-    def update(self, event: Event):
+    def update(self, event: Event) -> None:
         """update `event` in vdir and db"""
-        assert event.etag
+        assert event.etag is not None
+        assert event.calendar is not None
+        assert event.href is not None
+        assert event.raw is not None
         if self._calendars[event.calendar]['readonly']:
             raise ReadOnlyCalendarError()
         with self._backend.at_once():
@@ -172,9 +182,11 @@ class CalendarCollection(object):
             self._backend.update(event.raw, event.href, event.etag, calendar=event.calendar)
             self._backend.set_ctag(self._local_ctag(event.calendar), calendar=event.calendar)
 
-    def force_update(self, event: Event, collection: Optional[str]=None):
+    def force_update(self, event: Event, collection: Optional[str]=None) -> None:
         """update `event` even if an event with the same uid/href already exists"""
+        href: str
         calendar = collection if collection is not None else event.calendar
+        assert calendar is not None
         if self._calendars[calendar]['readonly']:
             raise ReadOnlyCalendarError()
 
@@ -188,21 +200,26 @@ class CalendarCollection(object):
             self._backend.update(event.raw, href, etag, calendar=calendar)
             self._backend.set_ctag(self._local_ctag(calendar), calendar=calendar)
 
-    def new(self, event: Event, collection: Optional[str]=None):
-        """save a new event to the vdir and the database
+    def insert(self, event: Event, collection: Optional[str]=None) -> None:
+        """Insert a new event to the vdir and the database
 
-        param event: the event that should be updated, will get a new href and
-            etag properties
-        type event: event.Event
+        The event will get a new href and etag properties. If ``collection`` is
+        ``None``, then ``event.calendar`` must be defined.
+
+        :param event: the event to be inserted.
         """
+        # TODO FIXME not all `event`s are actually of type Event, we also uptade
+        # with vdir.Items. Those don't have an .href or .etag property which we
+        # than attach anyway. Works, but pretty ugly and any type checker will
+        # complain.
         calendar = collection if collection is not None else event.calendar
+        assert calendar is not None
         if hasattr(event, 'etag'):
             assert not event.etag
         if self._calendars[calendar]['readonly']:
             raise ReadOnlyCalendarError()
 
         with self._backend.at_once():
-
             try:
                 event.href, event.etag = self._storages[calendar].upload(event)
             except AlreadyExistingError as Error:
@@ -211,27 +228,51 @@ class CalendarCollection(object):
             self._backend.update(event.raw, event.href, event.etag, calendar=calendar)
             self._backend.set_ctag(self._local_ctag(calendar), calendar=calendar)
 
-    def delete(self, href: str, etag: str, calendar: str):
+    def delete(self, href: str, etag: Optional[str], calendar: str) -> None:
+        """Delete an event specified by `href` from `calendar`"""
         if self._calendars[calendar]['readonly']:
             raise ReadOnlyCalendarError()
-        self._storages[calendar].delete(href, etag)
+        try:
+            self._storages[calendar].delete(href, etag)
+        except WrongEtagError:
+            raise EtagMissmatch()
         self._backend.delete(href, calendar=calendar)
 
+    def delete_instance(self,
+                        href: str,
+                        etag: Optional[str],
+                        calendar: str,
+                        rec_id: dt.datetime,
+                        ) -> Event:
+        """Delete a recurrence instance from an event specified by `href` from `calendar`
+
+        returns the updated event
+        """
+        if self._calendars[calendar]['readonly']:
+            raise ReadOnlyCalendarError()
+        event = self.get_event(href, calendar)
+        if etag and etag != event.etag:
+            raise EtagMissmatch()
+
+        event.delete_instance(rec_id)
+        self.update(event)
+        return event
+
     def get_event(self, href: str, calendar: str) -> Event:
         """get an event by its href from the datatbase"""
-        return self._construct_event(
-            self._backend.get(href, calendar), href=href, calendar=calendar,
-        )
+        event_str, etag = self._backend.get_with_etag(href, calendar)
+        return self._construct_event(event_str, etag=etag, href=href, calendar=calendar)
 
     def _construct_event(self,
                          item: str,
                          href: str,
-                         start: dt.datetime = None,
-                         end: dt.datetime = None,
+                         start: Optional[Union[dt.datetime, dt.date]] = None,
+                         end: Optional[Union[dt.datetime, dt.date]] = None,
                          ref: str='PROTO',
-                         etag: str=None,
-                         calendar: str=None,
+                         etag: Optional[str]=None,
+                         calendar: Optional[str]=None,
                          ) -> Event:
+        assert calendar is not None
         event = Event.fromString(
             item,
             locale=self._locale,
@@ -246,19 +287,38 @@ class CalendarCollection(object):
         )
         return event
 
-    def change_collection(self, event: Event, new_collection: str):
+    def change_collection(self, event: Event, new_collection: str) -> None:
+        """Moves `event` to a new collection (calendar)"""
         href, etag, calendar = event.href, event.etag, event.calendar
         event.etag = None
-        self.new(event, new_collection)
+        self.insert(event, new_collection)
+        assert href is not None
+        assert calendar is not None
         self.delete(href, etag, calendar=calendar)
 
-    def new_event(self, ical: str, collection: str):
-        """creates and returns (but does not insert) new event from ical
+    def create_event_from_ics(self,
+                              ical: str,
+                              calendar_name: str,
+                              etag: Optional[str]=None,
+                              href: Optional[str]=None,
+                              ) -> Event:
+        """creates and returns (but does not insert) a new event from ical
         string"""
-        calendar = collection or self.writable_names[0]
-        return Event.fromString(ical, locale=self._locale, calendar=calendar)
+        calendar = calendar_name or self.writable_names[0]
+        return Event.fromString(ical, locale=self._locale, calendar=calendar, etag=etag, href=href)
 
-    def update_db(self):
+    def create_event_from_dict(self,
+                               event_dict: EventCreationTypes,
+                               calendar_name: Optional[str] = None,
+                               ) -> Event:
+        """Creates an Event from the method's arguments
+        """
+        vevent = new_vevent(locale=self._locale, **event_dict)
+        calendar_name = calendar_name or self.default_calendar_name or self.writable_names[0]
+        assert calendar_name is not None
+        return self.create_event_from_ics(vevent.to_ical(), calendar_name)
+
+    def update_db(self) -> None:
         """update the db from the vdir,
 
         should be called after every change to the vdir
@@ -300,11 +360,11 @@ class CalendarCollection(object):
             self._last_ctags[calendar] = local_ctag
         return local_ctag != self._backend.get_ctag(calendar)
 
-    def _db_update(self, calendar: str):
+    def _db_update(self, calendar: str) -> None:
         """implements the actual db update on a per calendar base"""
         local_ctag = self._local_ctag(calendar)
-        db_hrefs = set(href for href, etag in self._backend.list(calendar))
-        storage_hrefs = set()
+        db_hrefs = {href for href, etag in self._backend.list(calendar)}
+        storage_hrefs: Set[str] = set()
         bdays = self._calendars[calendar].get('ctype') == 'birthdays'
 
         with self._backend.at_once():
@@ -312,7 +372,7 @@ class CalendarCollection(object):
                 storage_hrefs.add(href)
                 db_etag = self._backend.get_etag(href, calendar=calendar)
                 if etag != db_etag:
-                    logger.debug('Updating {0} because {1} != {2}'.format(href, etag, db_etag))
+                    logger.debug(f'Updating {href} because {etag} != {db_etag}')
                     self._update_vevent(href, calendar=calendar)
             for href in db_hrefs - storage_hrefs:
                 if bdays:
@@ -341,8 +401,8 @@ class CalendarCollection(object):
             if not isinstance(e, (UpdateFailed, UnsupportedFeatureError, NonUniqueUID)):
                 logger.exception('Unknown exception happened.')
             logger.warning(
-                'Skipping {0}/{1}: {2}\n'
-                'This event will not be available in khal.'.format(calendar, href, str(e)))
+                f'Skipping {calendar}/{href}: {str(e)}\n'
+                'This event will not be available in khal.')
             return False
 
     def search(self, search_string: str) -> Iterable[Event]:
@@ -357,11 +417,11 @@ class CalendarCollection(object):
             return 'highlight_days_color'
         if len(calendars) == 1:
             return 'calendar ' + calendars[0]
-        if self.multiple != '':
+        if self.multiple != '' and not (self.multiple_on_overflow and len(calendars) == 2):
             return 'highlight_days_multiple'
         return ('calendar ' + calendars[0], 'calendar ' + calendars[1])
 
-    def get_styles(self, date: dt.date, focus: bool) -> Union[str, None, Tuple[str, str]]:
+    def get_styles(self, date: dt.date, focus: bool) -> Optional[Union[str, Tuple[str, str]]]:
         if focus:
             if date == date.today():
                 return 'today focus'
